import os

import click
from sqlalchemy import create_engine

from dinero import analysis


@click.command()
def mkdataset():
    """Export all transactions to CSV and SQLite."""
    if not os.path.exists("data"):
        os.makedirs("data", exist_ok=True)

    file_path = "data/all.csv"

    df = analysis.get_dataframe()
    df.to_csv(file_path, index=False)

    engine = create_engine("sqlite:///data/transactions.db")

    with engine.connect() as conn:
        df.to_sql("transactions", conn, if_exists="replace", index=False)



import json
import os

import click
from loguru import logger
from sqlalchemy import select

from dinero.application import Application
from dinero.db import Transaction, get_session


@click.command()
def cache():
    """Cache accounts, categories and subcategories.

    Creates JSON files in ~/.config/dinero/cache/ that can be used by
    AI agents or other tools to know valid filter values without querying
    the database directly.
    """
    app = Application()
    session = get_session(app)

    cache_dir = app.cache_dir
    os.makedirs(cache_dir, exist_ok=True)

    # Query distinct accounts
    stmt = (
        select(Transaction.account)
        .where(Transaction.account.isnot(None))
        .where(Transaction.account != "")
        .distinct()
        .order_by(Transaction.account)
    )
    accounts = [row for row in session.scalars(stmt)]

    # Query distinct categories
    stmt = (
        select(Transaction.category)
        .where(Transaction.category.isnot(None))
        .where(Transaction.category != "")
        .distinct()
        .order_by(Transaction.category)
    )
    categories = [row for row in session.scalars(stmt)]

    # Query distinct subcategories
    stmt = (
        select(Transaction.subcategory)
        .where(Transaction.subcategory.isnot(None))
        .where(Transaction.subcategory != "")
        .distinct()
        .order_by(Transaction.subcategory)
    )
    subcategories = [row for row in session.scalars(stmt)]

    # Query distinct (category, subcategory) pairs for richer context
    stmt = (
        select(Transaction.category, Transaction.subcategory)
        .where(Transaction.category.isnot(None))
        .where(Transaction.category != "")
        .where(Transaction.subcategory.isnot(None))
        .where(Transaction.subcategory != "")
        .distinct()
        .order_by(Transaction.category, Transaction.subcategory)
    )
    category_pairs = [
        {"category": row[0], "subcategory": row[1]} for row in session.execute(stmt)
    ]

    # Write accounts.json
    accounts_path = cache_dir / "accounts.json"
    with open(accounts_path, "w") as f:
        json.dump(accounts, f, indent=2)
    logger.info(f"Cached {len(accounts)} accounts to {accounts_path}")

    # Write categories.json
    categories_data = {
        "categories": categories,
        "subcategories": subcategories,
        "category_subcategory_pairs": category_pairs,
    }
    categories_path = cache_dir / "categories.json"
    with open(categories_path, "w") as f:
        json.dump(categories_data, f, indent=2)
    logger.info(
        f"Cached {len(categories)} categories and {len(subcategories)} subcategories to {categories_path}"
    )

    session.close()

    logger.info(f"Cache written to {cache_dir}")



from typing import Dict

from pydantic import BaseModel, BaseSettings


class Plaid(BaseModel):
    client_id: str
    secret: str
    env: str
    products: str
    tokens: Dict[str, str]
    account_id_to_name: Dict[str, str]


class Database(BaseModel):
    connection_string: str


class RootConfig(BaseSettings):
    plaid: Plaid
    database: Database

    timezone: str = "US/Central"

    class Config:
        env_prefix = "dinero_"



from typing import Union

from dinero.config.config_file import ConfigFile
from dinero.config.model import RootConfig
from dinero.utils.fs import Path


class Application:
    config_file_path: Union[Path, None]
    config_file: ConfigFile

    def __init__(
        self,
        config_file: Union[str, None] = None,  # Path to config file
        load_config=True,
    ):
        if config_file is None:
            self.config_file_path = ConfigFile.get_default_location()
            self.config_file = ConfigFile()
        else:
            self.config_file_path = Path(config_file)
            path_ = None if config_file is None else Path(config_file)
            self.config_file = ConfigFile(path=path_)

        if load_config:
            self.load_config()

    def load_config(self):
        self.config_file.load()

    @property
    def config(self) -> RootConfig:
        return self.config_file.model

    @property
    def config_dir(self):
        return self.config_file.configdir()

    @property
    def cache_dir(self):
        return self.config_file.configdir() / "cache"


Dinero = Application



import datetime

from numpy import select
from loguru import logger
from sqlalchemy import DateTime, Double, String, create_engine, select
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, sessionmaker

from dinero import rules
from dinero.application import Application


class Table:
    def __init__(self, app: Application):
        self.app = app
        self.session = get_session(app)

        # Internal state
        self.records = []
        self.new = []
        self.existing = []
        self.errored = []

        self.load()

    def __repr__(self):
        return "Transactions: {}".format(len(self.records))

    def __len__(self):
        return len(self.records)

    def load(self):
        stmt = select(Transaction)

        for transaction in self.session.scalars(stmt):
            self.records.append(transaction)

    def prepare(self, transactions):
        """
        Checks:
            1. If the records to be inserted exist
            2. If its the correct table

        Parameters
        ----------
            transactions: PlaidTransactions with new records to (possibly) be inserted

        Returns
        -------
            (existing, new, error): tuple of 3 lists, each list are Record
        """
        new_records = [Transaction.from_plaid(self.app, tr) for tr in transactions]

        self.new, self.existing = [], []

        for record in new_records:
            if not self.exists(record):
                self.new.append(record)
            else:
                self.existing.append(record)
            # break

        return self.new, self.existing

    def prepare_from_records(self, records: list["Transaction"]):
        """
        Prepare Transaction records for insertion (for CSV import).

        Unlike `prepare()` which converts from Plaid format, this takes
        already-created Transaction objects.

        Parameters
        ----------
            records: List of Transaction objects to (possibly) be inserted

        Returns
        -------
            (new, existing): tuple of 2 lists of Transaction objects
        """
        self.new, self.existing = [], []

        for record in records:
            if not self.exists(record):
                self.new.append(record)
            else:
                self.existing.append(record)

        return self.new, self.existing

    def commit(self, *args, **kwargs):
        """Commit changes to Postgres
        Run `self.prepare(records)` first
        """
        self.session.add_all(self.new)
        self.session.commit()

    def exists(self, new):
        """Check if the record exists"""
        for existing in self.records:
            if existing == new:
                return True
        return False

    def close(self):
        self.session.close()


def get_session(app: Application):
    engine = create_engine(app.config.database.connection_string)

    Session = sessionmaker(bind=engine)
    return Session()


class Base(DeclarativeBase):
    pass


class Transaction(Base):
    """The transactions table

    Usage
    -----
        table.prepare(new_records)
        table.commit()
    """

    __tablename__ = "transactions"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    date: Mapped[datetime.date] = mapped_column(DateTime(timezone=True), nullable=True)
    description: Mapped[str] = mapped_column(String(1000), nullable=True)
    category: Mapped[str] = mapped_column(String(50), nullable=True)
    subcategory: Mapped[str] = mapped_column(String(50), nullable=True)
    amount: Mapped[float] = mapped_column(Double(), nullable=True)
    notes: Mapped[str] = mapped_column(String(1000), nullable=True)
    account: Mapped[str] = mapped_column(String(50), nullable=True)

    def __eq__(self, other_record):
        # Checks uniqueness based on: Date, Account Name, Description, Amount
        my_date = "%s-%02d-%02d" % (
            self.date.year,
            self.date.month,
            self.date.day,
        )
        date = "%s-%02d-%02d" % (
            other_record.date.year,
            other_record.date.month,
            other_record.date.day,
        )

        if (
            my_date == date
            and self.account == other_record.account
            and self.description == other_record.description
            and self.amount == other_record.amount
        ):
            return True
        return False

    @classmethod
    def from_plaid(cls, app: Application, plaid_transaction):
        new = cls()
        description = plaid_transaction.name
        category, subcategory = "", ""
        category, subcategory = rules.categories_for_transaction(app, description)

        new.date = plaid_transaction.date
        new.description = description
        new.category = category
        new.subcategory = subcategory
        new.amount = plaid_transaction.amount
        new.notes = "new"
        new.account = plaid_transaction.account_name

        return new

    def __repr__(self):
        return f"Transaction({self.id}): {self.date} {self.description} {self.amount}"



import os
import json

from loguru import logger
from dinero.application import Application


def categories_for_transaction(app: Application, description):
    """Returns category and subcategory for a transaction description
    based on the existing rules
    """

    generated_rules_path = app.config_dir / "category_rules.json"

    if os.path.exists(generated_rules_path):
        with open(generated_rules_path, "r") as fp:
            rules = json.load(fp)
    else:
        rules = []

    if description in rules:
        cat, subcat = rules[description][0], rules[description][1]
        logger.bind(category=cat, subcategory=subcat, desc=description).debug(
            "Automatically adding categories"
        )

        return cat, subcat
    return "", ""



from __future__ import annotations

import os
import pathlib
import sys
from typing import Any

import tomllib

repo_root = pathlib.Path(__file__).parent.parent.parent.resolve()


disk_sync = os.fsync
# https://mjtsai.com/blog/2022/02/17/apple-ssd-benchmarks-and-f_fullsync/
# https://developer.apple.com/library/archive/documentation/System/Conceptual/ManPages_iPhoneOS/man2/fsync.2.html
if sys.platform == "darwin":
    import fcntl

    if hasattr(fcntl, "F_FULLFSYNC"):

        def disk_sync(fd) -> None:
            fcntl.fcntl(fd, fcntl.F_FULLFSYNC)


class Path(type(pathlib.Path())):
    def __new__(cls, *pathsegments):
        return super().__new__(cls, *pathsegments)

    def ensure_dir_exists(self) -> None:
        self.mkdir(parents=True, exist_ok=True)

    def ensure_parent_dir_exists(self) -> None:
        self.parent.mkdir(parents=True, exist_ok=True)

    def expand(self) -> Path:
        return Path(os.path.expanduser(os.path.expandvars(self)))

    def resolve(self, strict: bool = False) -> Path:  # noqa: FBT001, FBT002
        # https://bugs.python.org/issue38671
        return Path(os.path.realpath(self))

    def remove(self) -> None:
        if self.is_file():
            os.remove(self)
        elif self.is_dir():
            import shutil

            shutil.rmtree(self, ignore_errors=False)

    def write(self, data: str) -> None:
        self.parent.mkdir(parents=True, exist_ok=True)
        with self.open("w") as f:
            f.write(data)

    def write_atomic(self, data: str | bytes, *args: Any, **kwargs: Any) -> None:
        from tempfile import mkstemp

        fd, path = mkstemp(dir=self.parent)
        with os.fdopen(fd, *args, **kwargs) as f:
            f.write(data)
            f.flush()
            disk_sync(fd)

        os.replace(path, self)


def load_toml_data(data: str) -> dict[str, Any]:
    return tomllib.loads(data)


def load_toml_file(path: str) -> dict[str, Any]:
    with open(path, encoding="utf-8") as f:
        return tomllib.loads(f.read())



import click
from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy_utils import create_database, database_exists

from dinero import Application
from dinero.cli import utils
from dinero.db import Transaction


@click.command()
def init_db():
    """Create the database and tables."""
    app = Application()

    conn_string = app.config.database.connection_string

    engine = create_engine(conn_string)

    if not database_exists(engine.url):
        if utils.noninteractive() or utils.query_yes_no(
            f"Create database: {conn_string}"
        ):
            create_database(engine.url)
        print("Database created")
    else:
        print("Database already exists. Skipping creation.")

    if utils.noninteractive() or utils.query_yes_no("Create tables?"):
        Transaction.metadata.create_all(engine)
        print("Table created")



"""Import transactions from a CSV file into the database.

CSV format expected:
    date,description,amount[,category,subcategory]

Example:
    date,description,amount
    2026-01-31,Deposits,5000.00
    2026-01-31,Withdrawals,-1200.00

Usage:
    dinero import-csv ./transactions.csv --account "Brokerage Account"
"""

import csv
import sys
from pathlib import Path

import click
import pendulum
from loguru import logger

from dinero import Application, db, rules
from dinero.cli import utils


@click.command()
@click.argument("file")
@click.argument("account")
def import_csv(file: str, account: str):
    """Import transactions from a CSV file.

    FILE is the path to the CSV file to import.

    ACCOUNT is the account name to associate with all transactions.
    """
    app = Application()
    csv_path = Path(file)

    if not csv_path.exists():
        logger.error(f"File not found: {csv_path}")
        sys.exit(1)

    # Parse CSV file
    transactions = parse_csv(app, csv_path, account)

    if not transactions:
        logger.warning("No transactions found in CSV file")
        return

    # Load existing transactions and check for duplicates
    table = db.Table(app)
    new, existing = table.prepare_from_records(transactions)

    # Display summary
    print("=" * 80)
    print(f"Import from: {csv_path}")
    print(f"Account: {account}")
    print("=" * 80)
    print(f"Transactions in CSV: {len(transactions)}")
    print(f"New transactions: {len(new)}")
    print(f"Already existing (will skip): {len(existing)}")
    print()

    if not new:
        print("No new transactions to import.")
        return

    # Show transactions to be imported
    print("-" * 80)
    print("Transactions to be imported:")
    print("-" * 80)
    for t in new:
        category_str = f" [{t.category}]" if t.category else ""
        print(
            f"  {t.date.strftime('%Y-%m-%d')}  {t.amount:>10.2f}  {t.description}{category_str}"
        )
    print()

    # Confirm and commit
    if utils.noninteractive() or utils.query_yes_no("Import these transactions?"):
        print(f"Inserting {len(new)} new records...")
        table.commit()
        print("Done")
    else:
        print("Import cancelled.")


def parse_csv(app: Application, csv_path: Path, account: str) -> list[db.Transaction]:
    """Parse a CSV file and return a list of Transaction objects.

    Expected columns: date, description, amount
    Optional columns: category, subcategory
    """
    transactions = []

    with open(csv_path, newline="", encoding="utf-8") as f:
        reader = csv.DictReader(f)

        # Validate required columns
        required_columns = {"date", "description", "amount"}
        if not required_columns.issubset(set(reader.fieldnames or [])):
            missing = required_columns - set(reader.fieldnames or [])
            logger.error(f"Missing required columns: {missing}")
            logger.info("Required columns: date, description, amount")
            sys.exit(1)

        for row_num, row in enumerate(
            reader, start=2
        ):  # start=2 because row 1 is header
            try:
                transaction = parse_row(app, row, account, row_num)
                if transaction:
                    transactions.append(transaction)
            except ValueError as e:
                logger.error(f"Row {row_num}: {e}")
                sys.exit(1)

    return transactions


def parse_row(
    app: Application, row: dict, account: str, row_num: int
) -> db.Transaction | None:
    """Parse a single CSV row into a Transaction object."""
    date_str = row.get("date", "").strip()
    description = row.get("description", "").strip()
    amount_str = row.get("amount", "").strip()

    # Skip empty rows
    if not date_str and not description and not amount_str:
        return None

    # Validate required fields
    if not date_str:
        raise ValueError("Missing date")
    if not description:
        raise ValueError("Missing description")
    if not amount_str:
        raise ValueError("Missing amount")

    # Parse date (ISO format: YYYY-MM-DD)
    try:
        date = pendulum.parse(date_str, tz=app.config.timezone)
    except Exception:
        raise ValueError(
            f"Invalid date format: '{date_str}'. Expected ISO format (YYYY-MM-DD)"
        )

    # Parse amount
    try:
        amount = float(amount_str.replace(",", ""))
    except ValueError:
        raise ValueError(f"Invalid amount: '{amount_str}'")

    # Get category from CSV or from rules
    category = row.get("category", "").strip()
    subcategory = row.get("subcategory", "").strip()

    if not category:
        category, subcategory = rules.categories_for_transaction(app, description)

    # Create Transaction object
    t = db.Transaction()
    t.date = date
    t.description = description
    t.amount = amount
    t.category = category
    t.subcategory = subcategory
    t.notes = "csv-import"
    t.account = account

    return t



import sys

import click
from loguru import logger

from dinero import Application
from dinero.cli.cache import cache
from dinero.cli.db import init_db
from dinero.cli.import_csv import import_csv
from dinero.cli.mkdataset import mkdataset
from dinero.cli.mkrules import gen_rules
from dinero.cli.search import search
from dinero.cli.transactions import transactions


# Add a new handler that only shows INFO and above
logger.remove()
logger.add(
    sys.stdout,
    level="INFO",
    colorize=True,
    format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | {message} | <cyan>{extra}</cyan>",
)


@click.group()
def main():
    """Dinero - Personal finance tracking tools."""
    pass


@main.command("config")
def config_cmd():
    """Print the config file path and loaded configuration."""
    app = Application()

    print(app.config_file_path)
    print(app.config)


main.add_command(init_db, "init-db")
main.add_command(mkdataset, "mkdataset")
main.add_command(transactions, "transactions")
main.add_command(gen_rules, "mkrules")
main.add_command(import_csv, "import-csv")
main.add_command(search, "search")
main.add_command(cache, "build-cache")


if __name__ == "__main__":
    main()



# A dashboard that shows the balance for all the accounts from accounts.yml

import streamlit as st
import pandas as pd
import yaml
from datetime import datetime, time

from dinero import analysis
from dinero.utils.base import today, this_or_last_month


def to_utc(dt):
    dt = pd.to_datetime(dt)
    if dt.tzinfo is None:
        dt = dt.tz_localize("UTC")
    return dt


# Load configuration from YAML file
with open("accounts.yml", "r") as f:
    config = yaml.safe_load(f)

df = analysis.get_dataframe()

accounts = {}

# Create a sidebar date input for each account based on configuration
for account in config["accounts"]:
    name = account["name"]
    default_config = account["default"]

    if default_config["method"] == "today":
        default_date = today
    elif default_config["method"] == "this_or_last_month":
        offset = default_config.get("offset", 0)
        default_date = this_or_last_month(offset)
    else:
        default_date = today  # fallback

    user_date = st.sidebar.date_input(name, default_date)

    # Combine the selected date with the end-of-day time to make the filter inclusive
    end_of_day_datetime = datetime.combine(user_date, time(23, 59, 59))
    accounts[name] = to_utc(end_of_day_datetime)

# Process each account's data
values = []
for account_name, account_date in accounts.items():
    # The 'before' parameter will now correctly include the entire selected day
    data = analysis.select(df, account=account_name, before=account_date)
    balance = "{:,.2f}".format(data.amount.sum())
    values.append(
        {"Account": account_name, "Date": account_date.date(), "Balance": balance}
    )

accounts_df = pd.DataFrame(values).set_index("Account")

st.write("# Balances per Date")
st.dataframe(accounts_df, use_container_width=True)



import pandas as pd
from loguru import logger

from dinero import Application
from dinero.db import Transaction

pd.options.display.float_format = "{:,.2f}".format


app = Application()


def get_dataframe():
    """Get data from the DB and reteurn a pandas.DataFrame"""
    con = app.config.database.connection_string
    df = pd.read_sql(Transaction.__tablename__, con=con)
    return df


def select(data=None, year=None, month=None, before=None, after=None, account=None):
    selected = data if data is not None else get_dataframe()

    # Convert only if the parameter is provided and is not already null.
    before = pd.to_datetime(before) if before is not None else None
    after = pd.to_datetime(after) if after is not None else None

    if account:
        selected = selected[selected.account == account]
    if year:
        selected = selected[selected.date.dt.year == year]
    if month:
        selected = selected[selected.date.dt.month == month]
    if after is not None and pd.notnull(after):
        selected = selected[selected.date >= after]
    if before is not None and pd.notnull(before):
        selected = selected[selected.date <= before]
    return selected


def group_desc_categories(data):
    """Group the data by
    Useful to see the most common transactions and generate rules
    """
    groups = data.groupby(["description", "category", "subcategory"]).count()
    groups = groups.amount.sort_values(ascending=False)
    return groups[groups >= 3]



import pendulum
from loguru import logger
from pendulum import Date
from plaid import Client
from plaid import errors as plaid_errors

from dinero.application import Application
from dinero.utils import base as baseutils


CLIENT = None


def get_client(app: Application):
    """Get a Plaid client"""
    global CLIENT
    if not CLIENT:
        plaid_client_id = app.config.plaid.client_id
        plaid_secret = app.config.plaid.secret
        # plaid_public_key = app.config.plaid.public_key
        plaid_env = app.config.plaid.env

        CLIENT = Client(
            client_id=plaid_client_id,
            secret=plaid_secret,
            # public_key=plaid_public_key,
            environment=plaid_env,
            suppress_warnings=True,
        )

    return CLIENT


def get_all_transactions(app: Application, date: str | Date, days=30):
    """Get transactions for all institutions"""
    all_transactions = TransactionList()
    for account_name, access_token in app.config.plaid.tokens.items():
        all_transactions.extend(
            get_transactions(app, account_name, date=date, days=days)
        )
    return all_transactions


def get_transactions(app: Application, name, date: str | Date, days=30):
    """Get transactions for one account based on the access token

    Parameters
    ----------
        name: Name of the institution on the config
        date: last day to end the transactions range
        days: number of days back from `date` to get transactions for
    """
    access_token = app.config.plaid.tokens[name]
    start_date, end_date = baseutils.get_dates_from_delta(date=date, days=days)

    # Get transactions in Plaid API JSON
    client = get_client(app)
    try:
        response = client.Transactions.get(
            access_token, start_date=start_date, end_date=end_date
        )
    except plaid_errors.ItemError as ex:
        if "the login details of this item have changed" in str(ex):
            raise Exception("Credentials expired for account: %s" % name)
        else:
            raise ex

    # This loop goes through the Plaid pagination
    transactions_api_json = response["transactions"]
    while len(transactions_api_json) < response["total_transactions"]:
        response = client.Transactions.get(
            access_token,
            start_date=start_date,
            end_date=end_date,
            offset=len(transactions_api_json),
        )
        transactions_api_json.extend(response["transactions"])
    logger.bind(name=name, records=len(transactions_api_json)).info(
        "Transactions downloaded"
    )

    # From Plaid API JSON format to the classes in this file
    transactions = TransactionList()
    transactions.app = app

    for transaction_json in transactions_api_json:
        transactions.append_from_plaid(transaction_json)

    logger.bind(name=name, records=len(transactions.pending)).info(
        "Transactions pending"
    )
    logger.bind(name=name, records=len(transactions.not_pending)).info(
        "Transactions not pending"
    )
    return transactions


def account_id_to_name(app: Application, account_id):
    """Convert a Plaid account ID to a human readable name from the config"""
    plaid_id_to_name = app.config.plaid.account_id_to_name
    if account_id in plaid_id_to_name:
        return plaid_id_to_name[account_id]
    else:
        logger.error(
            "ID not found in the accounts map, add it to the settings.toml "
            "on section: plaid.account_id_to_name",
            account_id=account_id,
        )
        return account_id


class TransactionList(list):
    app: Application

    @property
    def pending(self):
        """Filter by pending transactions
        Returns a new TransactionList
        """
        return TransactionList(filter(lambda x: x.pending, self))

    @property
    def not_pending(self):
        """Filter by not pending transactions
        Returns a new TransactionList
        """
        return TransactionList(filter(lambda x: not x.pending, self))

    def append(self, item, *args, **kwargs):
        if isinstance(item, Transaction):
            super(TransactionList, self).append(item, *args, **kwargs)
        else:
            raise ValueError(
                "TransactionsList class can only append Transaction objects"
            )

    def append_from_plaid(self, transaction_json):
        """Convert from plaid json to Transaction class and append"""
        new = Transaction.from_plaid(self.app, transaction_json)
        self.append(new)

    def group_by_account(self):
        """Returns a dictionary of {account_name: transaction_of_that_account}"""
        groups = {}
        for item in self:
            account_name = item.account_name
            if account_name in groups:
                groups[account_name].append(item)
            else:
                t_list = TransactionList()
                t_list.append(item)
                groups[account_name] = t_list
        return groups


class Transaction(object):
    def __init__(self, app):
        self.app: Application = app
        self.account_id = ""
        self.amount = 0
        self.category = []
        self.category_id = 0
        self.date: Date = pendulum.now()
        self.name = ""
        self.pending = False

    @classmethod
    def from_plaid(cls, app: Application, values):
        """Create a class from plaid API transaction JSON"""
        new = cls(app)
        new.account_id = values["account_id"]
        new.amount = values["amount"]
        new.category = values["category"]
        new.category_id = values["category_id"]
        new.date = pendulum.parse(values["date"], tz=app.config.timezone)
        new.name = values["name"]
        new.pending = values["pending"]
        return new

    @property
    def account_name(self):
        return account_id_to_name(self.app, self.account_id)

    def __str__(self):
        return str(
            {
                "account_name": self.account_name,
                "description": self.name,
                "date": self.date.strftime("%Y-%m-%d"),
                "amount": self.amount,
            }
        )

    def __repr__(self):
        return str(self.__dict__)



import sys
import os
from pathlib import Path
from typing import Union, cast

from dinero.config.model import RootConfig
from dinero.utils.fs import Path, load_toml_data

from platformdirs import PlatformDirs as ResultPD

if sys.platform == "darwin":
    from platformdirs.unix import Unix as ResultPD


class ConfigFile:
    model: RootConfig

    def __init__(self, path: Union[Path, None] = None):
        self._path: Union[Path, None] = path
        self.model = cast(RootConfig, None)

    @property
    def path(self):
        if self._path is None:
            self._path = self.get_default_location()
        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    def save(self, content=None):
        import tomli_w

        if not content:
            content = tomli_w.dumps(self.model.dict(exclude_none=True))

        self.path.ensure_parent_dir_exists()
        self.path.write_atomic(content, "w", encoding="utf-8")

    def load(self):
        self.model = RootConfig.parse_obj(load_toml_data(self.read()))

    def read(self) -> str:
        return self.path.read_text("utf-8")

    # def restore(self):
    #     import tomli_w

    #     config = RootConfig()

    #     content = tomli_w.dumps(config.dict(exclude_none=True))
    #     self.save(content)

    #     self.model = config

    def update(self):
        self.save()

    @classmethod
    def configdir(cls) -> Path:
        platform_dir = ResultPD("dinero")
        return Path(platform_dir.user_config_dir)

    @classmethod
    def get_default_location(cls) -> Path:
        key = "DINERO_CONFIG_FILE"
        env_value = os.getenv(key)
        if env_value is not None and os.path.exists(env_value):
            return Path(env_value)

        return cls.configdir() / "config.toml"



import click
import pendulum
from loguru import logger

from dinero import Application, db, plaid
from dinero.cli import utils
from dinero.utils import base as baseutils


DATE = pendulum.now()
# Explicitly set date
# DATE = pendulum.datetime(2022, 12, 31)
DAYS = 90
ADD_PENDING = False


@click.command()
def transactions():
    """Fetch new transactions from Plaid and insert into the database."""
    app = Application()

    # Get new transactions from plaid
    all_transactions = plaid.get_all_transactions(app, date=DATE, days=DAYS)

    table = db.Table(app)

    # Print individual accounts
    groups = all_transactions.group_by_account()
    for account_name, account_transactions in groups.items():
        print("-" * 80)
        print(account_name)
        print("-" * 80)

        new, existing = table.prepare(account_transactions.not_pending)

        n_transactions = len(account_transactions)
        n_pending = len(account_transactions.pending)
        n_new, n_existing = len(new), len(existing)

        logger.bind(n=n_transactions).info("Queried transactions")
        logger.bind(n=n_pending).info("Pending transactions")

        for transaction in account_transactions.pending:
            logger.info("Transaction", transaction=transaction)
        logger.bind(n=n_new).info("New transactions to be added")

        for transaction in new:
            logger.info("Transaction", transaction=transaction)
        logger.bind(n=n_existing).info("Existing transactions")

        # for transaction in existing:
        #     logger.info("Transaction", transaction=transaction)

        assert n_transactions == n_pending + n_new + n_existing
        print()

    # Print account summary
    new_records = all_transactions if ADD_PENDING else all_transactions.not_pending

    new, existing = table.prepare(new_records)

    len_table = len(table)
    n_transactions = len(all_transactions)
    n_records = len(new_records)
    n_pending = len(all_transactions.pending)
    n_new = len(new)
    n_existing = len(existing)

    start_date, end_date = baseutils.get_dates_from_delta(date=DATE, days=DAYS)

    print("=" * 80)
    print("All accounts summary from {} to {}:".format(start_date, end_date))
    print("=" * 80)
    print("Using Table for year: {}".format(DATE.year))
    print("Records in table: {}".format(len_table))
    print("Transactions queried: {}".format(n_transactions))
    print("Pending transactions: {}".format(n_pending))
    print("Transactions analysed (pending {}): {}".format(ADD_PENDING, n_records))
    print("New records to be inserted: {}".format(n_new))
    print("Existing transactions: {}".format(n_existing))
    print()

    assert n_records == n_new + n_existing

    if utils.noninteractive() or utils.query_yes_no(
        "Insert transactions to the Table?"
    ):
        print("Inserting %i new records" % len(table.new))
        table.commit()
        print("Done")



import json
import datetime

import click
from loguru import logger
from sqlalchemy import select, extract
from tabulate import tabulate

from dinero.application import Application
from dinero.db import Transaction, get_session


@click.command()
@click.option("--account", default=None, help="Filter by account name (exact match).")
@click.option("--category", default=None, help="Filter by category (exact match).")
@click.option(
    "--subcategory", default=None, help="Filter by subcategory (exact match)."
)
@click.option(
    "--description",
    default=None,
    help="Search description (case-insensitive partial match).",
)
@click.option(
    "--after", default=None, help="Transactions on or after this date (YYYY-MM-DD)."
)
@click.option(
    "--before", default=None, help="Transactions on or before this date (YYYY-MM-DD)."
)
@click.option("--year", type=int, default=None, help="Filter by year.")
@click.option("--month", type=int, default=None, help="Filter by month (1-12).")
@click.option(
    "--limit",
    type=int,
    default=50,
    show_default=True,
    help="Maximum number of rows to return.",
)
@click.option(
    "--sort",
    default="date",
    show_default=True,
    help="Column to sort by: date, amount, description, account, category.",
)
@click.option("--desc/--asc", default=True, show_default=True, help="Sort direction.")
@click.option(
    "--json",
    "json_output",
    is_flag=True,
    default=False,
    help="Output as JSON instead of table.",
)
def search(
    account,
    category,
    subcategory,
    description,
    after,
    before,
    year,
    month,
    limit,
    sort,
    desc,
    json_output,
):
    """Search transactions in the database with filters."""
    app = Application()
    session = get_session(app)

    stmt = select(Transaction)

    # Apply filters
    if account is not None:
        stmt = stmt.where(Transaction.account == account)

    if category is not None:
        stmt = stmt.where(Transaction.category == category)

    if subcategory is not None:
        stmt = stmt.where(Transaction.subcategory == subcategory)

    if description is not None:
        stmt = stmt.where(Transaction.description.ilike(f"%{description}%"))

    if after is not None:
        after_date = datetime.datetime.strptime(after, "%Y-%m-%d")
        stmt = stmt.where(Transaction.date >= after_date)

    if before is not None:
        before_date = datetime.datetime.strptime(before, "%Y-%m-%d")
        stmt = stmt.where(Transaction.date <= before_date)

    if year is not None:
        stmt = stmt.where(extract("year", Transaction.date) == year)

    if month is not None:
        stmt = stmt.where(extract("month", Transaction.date) == month)

    # Sorting
    sort_column_map = {
        "date": Transaction.date,
        "amount": Transaction.amount,
        "description": Transaction.description,
        "account": Transaction.account,
        "category": Transaction.category,
        "subcategory": Transaction.subcategory,
    }

    sort_col = sort_column_map.get(sort, Transaction.date)
    if desc:
        stmt = stmt.order_by(sort_col.desc())
    else:
        stmt = stmt.order_by(sort_col.asc())

    # Limit
    stmt = stmt.limit(limit)

    # Execute
    results = session.scalars(stmt).all()

    if not results:
        logger.info("No transactions found matching the given filters.")
        session.close()
        return

    # Format output
    rows = []
    for t in results:
        date_str = t.date.strftime("%Y-%m-%d") if t.date else ""
        rows.append(
            {
                "id": t.id,
                "date": date_str,
                "description": t.description or "",
                "category": t.category or "",
                "subcategory": t.subcategory or "",
                "amount": t.amount,
                "account": t.account or "",
                "notes": t.notes or "",
            }
        )

    if json_output:
        print(json.dumps(rows, indent=2))
    else:
        headers = [
            "id",
            "date",
            "description",
            "category",
            "subcategory",
            "amount",
            "account",
            "notes",
        ]
        table_data = [[row[h] for h in headers] for row in rows]
        print(tabulate(table_data, headers=headers, tablefmt="simple", floatfmt=",.2f"))

    logger.info(f"Found {len(results)} transaction(s)")
    session.close()



import calendar
import typing
from calendar import monthrange
from datetime import datetime

import pandas as pd
import pendulum
from pendulum import Date

today = datetime.now()
today = datetime(today.year, today.month, today.day, 23, 59, 59)
today = pd.to_datetime(today, utc=True)

def get_dates_from_month(date: str | Date):
    """
    Returns two strings with the first and end dates for a month
    """
    if isinstance(date, str):
        date = typing.cast(Date, pendulum.parse(date))
    date = date or pendulum.now()

    year, month = date.year, date.month
    first_weekday, last_day = monthrange(year, month)

    start_date = f"{year}-{month:02d}-01"
    end_date = f"{year}-{month:02d}-{last_day:02d}"
    return start_date, end_date


def get_dates_from_delta(date: str | Date, days=30):
    if isinstance(date, str):
        date = typing.cast(Date, pendulum.parse(date))
    end_date = date or pendulum.now()
    start_date = end_date.subtract(days=days)

    start_date = "{year}-{month:02d}-{day:02d}".format(
        year=start_date.year, month=start_date.month, day=start_date.day
    )
    end_date = "{year}-{month:02d}-{day:02d}".format(
        year=end_date.year, month=end_date.month, day=end_date.day
    )
    return start_date, end_date


def this_or_last_month(target_day=None):
    if target_day is None:
        target_day = today.day

    if target_day == -1:
        last_day_of_today_month = calendar.monthrange(today.year, today.month)[1]
        if today.day < last_day_of_today_month:
            if today.month == 1:
                target_month = 12
            else:
                target_month = today.month - 1
        else:
            target_month = today.month
        target_day = calendar.monthrange(today.year, target_month)[1]
    else:
        if target_day <= today.day:
            target_month = today.month
        else:
            if today.month == 1:
                target_month = 12
            else:
                target_month = today.month - 1

    if target_day >= today.day and target_month == 12:
        target_year = today.year - 1
    else:
        target_year = today.year

    dt = datetime(target_year, target_month, target_day, 23, 59, 59)
    return pd.to_datetime(dt, utc=True)



import json

import click

from dinero import Application, analysis
from dinero.cli import utils


@click.command()
def gen_rules():
    """Generate category rules from the most common transactions."""
    app = Application()

    TARGET = app.config_dir / "category_rules.json"

    data = analysis.get_dataframe()
    groups = analysis.group_desc_categories(data)
    most_common = groups[groups >= 3]
    most_common = most_common.reset_index()  # Remove MultiIndex

    print("Most common transactions:")
    print(most_common)

    if utils.noninteractive() or utils.query_yes_no(
        '\nDo you want to save these rules to "%s"?' % TARGET
    ):
        with open(TARGET, "w") as f:
            rules = {}
            for index, row in most_common.iterrows():
                rules[row["description"]] = [row["category"], row["subcategory"]]
            string = json.dumps(rules, sort_keys=True, indent=4, separators=(",", ": "))
            f.write(string)
